#!/usr/bin/python -tt
#
# edit-liveos: Edit a LiveOS to insert files
#
# Copyright 2009, Red Hat  Inc.
# Written by Perry Myers <pmyers at redhat.com> & David Huff <dhuff at redhat.com>
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import os
import sys
import stat
import tempfile
import shutil
import subprocess
import optparse
import logging

from imgcreate.debug import *
from imgcreate.fs import *
from imgcreate.live import *

class ExistingSparseLoopbackDisk(SparseLoopbackDisk):
    """don't want to expand the disk"""
    def __init__(self, lofile, size):
        SparseLoopbackDisk.__init__(self, lofile, size)

    def create(self):
        #self.expand(create = True)
        LoopbackDisk.create(self)

class LiveImageEditor(LiveImageCreator):
    """class for editing LiveOS images.

    We need an instance of LiveImageCreator however we do not have a kickstart
    file and we may not need to create a new image. We just want to reuse some of
    LiveImageCreators methods on an existing LiveOS image.

    """

    def __init__(self, name):
        """Initialize a LiveImageEditor instance.

        creates a dummy instance of LiveImageCreator
        We do not initialize any sub classes b/c we have no ks file.

        """
        self.name = name

        self.tmpdir = "/var/tmp"
        """The directory in which all temporary files will be created."""

        self.clone = False
        """Signals when to copy an attached LiveOS image as base."""

        self._include = None
        """A string of file or directory paths to include in __copy_img_root."""

        self._builder = "someone"
        """The name of the Remix builder for _branding."""

        self.skip_compression = False
        """Controls whether to use squashfs to compress the image."""

        self.skip_minimize = False
        """Controls whether an image minimizing snapshot should be created."""

        self._isofstype = "iso9660"
        self.__isodir = None

        self._ImageCreator__builddir = None
        """working directory"""

        self._ImageCreator_instroot = None
        """where the extfs.img is mounted for modification"""

        self._ImageCreator_outdir = None
        """where final iso gets written"""

        self._ImageCreator__bindmounts = []

        self._LoopImageCreator__imagedir = None
        """dir for the extfs.img"""

        self._LoopImageCreator__blocksize = 4096
        self._LoopImageCreator__fslabel = None
        self._LoopImageCreator__instloop = None
        self._LoopImageCreator__fstype = None
        self._LoopImageCreator__image_size = None

        self._LiveImageCreatorBase__isodir = None
        """directory where the iso is staged"""

    # properties
    def __get_image(self):
        if self._LoopImageCreator__imagedir is None:
            self.__ensure_builddir()
            self._LoopImageCreator__imagedir = \
                tempfile.mkdtemp(dir = os.path.abspath(self.tmpdir),
                                                       prefix = self.name + "-")
        if self.clone:
            rtn = self._LoopImageCreator__imagedir
        else:
            rtn = self._LoopImageCreator__imagedir + "/ext3fs.img"
        return rtn
    _image = property(__get_image)
    """The location of the image file"""


    def _get_fstype(self):
        dev_null = os.open("/dev/null", os.O_WRONLY)
        try:
            out = subprocess.Popen(["/sbin/blkid", self._image],
                                   stdout = subprocess.PIPE,
                                   stderr = dev_null).communicate()[0]
            for word in out.split():
                if word.startswith("TYPE"):
                    self._LoopImageCreator__fstype = word.split("=")[1].strip("\"")

        except IOError, e:
            raise CreatorError("Failed to determine fsimage TYPE: %s" % e )


    def _get_fslabel(self):
        dev_null = os.open("/dev/null", os.O_WRONLY)
        try:
            out = subprocess.Popen(["/sbin/e2label", self._image],
                                   stdout = subprocess.PIPE,
                                   stderr = dev_null).communicate()[0]

            self._LoopImageCreator__fslabel = out.strip()

        except IOError, e:
            raise CreatorError("Failed to determine fsimage TYPE: %s" % e )


    def __ensure_builddir(self):
        if not self._ImageCreator__builddir is None:
            return

        try:
            self._ImageCreator__builddir = tempfile.mkdtemp(dir =  os.path.abspath(self.tmpdir),
                                               prefix = "edit-liveos-")
        except OSError, (err, msg):
            raise CreatorError("Failed create build directory in %s: %s" %
                               (self.tmpdir, msg))


    def _run_script(self, script):

        (fd, path) = tempfile.mkstemp(prefix = "script-",
                                          dir = self._instroot + "/tmp")

        logging.debug("copying script to install root: %s" % path)
        shutil.copy(os.path.abspath(script), path)
        os.close(fd)
        os.chmod(path, 0700)

        script = "/tmp/" + os.path.basename(path)


        try:
            subprocess.call([script], preexec_fn = self._chroot)
        except OSError, e:
            raise CreatorError("Failed to execute script %s, %s " % (script, e))
        finally:
            os.unlink(path)

    def mount(self, base_on, cachedir = None):
        """mount existing file system.

        we have to override mount b/c we are not creating an new install root
        nor do we need to setup the file system, i.e., makedirs(/etc/, /boot, ...),
        nor do we want to overwrite fstab, or create selinuxfs

        We also need to get some info about the image before we
        can mount it.

        base_on --  the <LIVEIMG.src> a LiveOS.iso file or attached LiveOS device,
                    such as, /dev/live for a currently running image.

        cachedir -- a directory in which to store a Yum cache; Not used in edit-liveos.

        """

        if not base_on:
            raise CreatorError("No base LiveOS image specified.")

        self.__ensure_builddir()

        self._ImageCreator_instroot = self._ImageCreator__builddir + "/install_root"
        self._LoopImageCreator__imagedir = self._ImageCreator__builddir + "/ex"
        self._ImageCreator_outdir = self._ImageCreator__builddir + "/out"

        makedirs(self._ImageCreator_instroot)
        makedirs(self._LoopImageCreator__imagedir)
        makedirs(self._ImageCreator_outdir)

        if self.clone:
            self._base_on(base_on)
        else:
            LiveImageCreator._base_on(self, base_on)

        self._LoopImageCreator__image_size = os.stat(self._image)[stat.ST_SIZE]

        self._get_fstype()
        self._get_fslabel()
        self.fslabel = self._LoopImageCreator__fslabel

        if self.clone:
            self._LoopImageCreator__instloop = \
                ExtDiskMount(self._image,
                             self._LoopImageCreator__image_size,
                             self._ImageCreator_instroot,
                             self._fstype,
                             self._LoopImageCreator__blocksize,
                             self.fslabel)
        else:
            self._LoopImageCreator__instloop = \
                ExtDiskMount(ExistingSparseLoopbackDisk(self._image,
                                            self._LoopImageCreator__image_size),
                             self._ImageCreator_instroot,
                             self._fstype,
                             self._LoopImageCreator__blocksize,
                             self.fslabel)
        try:
            self._LoopImageCreator__instloop.mount()
        except MountError, e:
            raise CreatorError("Failed to loopback mount '%s' : %s" %
                               (self._image, e))

        cachesrc = cachedir or (self._ImageCreator__builddir + "/yum-cache")
        makedirs(cachesrc)

        for (f, dest) in [("/sys", None), ("/proc", None),
                          ("/dev/pts", None), ("/dev/shm", None),
                          (cachesrc, "/var/cache/yum")]:
            self._ImageCreator__bindmounts.append(BindChrootMount(f, self._instroot, dest))

        self._do_bindmounts()

        os.symlink("../proc/mounts", self._instroot + "/etc/mtab")

        self.__copy_img_root(base_on)

    def _base_on(self, base_on):
        """clone the running LiveOS image as the basis for the new image."""

#        imgmnt = DiskMount(RawDisk(base_on, 0), self._mkdtemp())
#                              , self._LoopImageCreator__imagedir)

#        try:
#            imgmnt.mount()

#        except MountError, e:
#            raise CreatorError("Failed to mount '%s' : %s" %
#                           (base_on, e))

        subprocess.call(["rsync", "-ptgorlHASx", "--specials", "--progress", 
                         "--exclude", "dev",
                         "--exclude", "proc",
                         "--exclude", "media",
                         "--exclude", "mnt",
                         "--exclude", "sys",
                         "--exclude", ".liveimg*",
                         "--exclude", "tmp",
                         "/", self._LoopImageCreator__imagedir])

    def _mount_instroot(self, base_on = None):
        self.__imgdir = self._mkdtemp()

        self.__instloop = ExtDiskMount(self._image,
                                       self._instroot,
                                       self.__fstype,
                                       self.__blocksize,
                                       self.fslabel)

        try:
            self.__instloop.mount()
        except MountError, e:
            raise CreatorError("Failed to mount '%s' : %s" %
                               (self._image, e))

    def __copy_img_root(self, base_on):
        """helper function to copy root content of the base LiveIMG to ISOdir"""

        ignore_list = ['squashfs.img', 'osmin.img', 'home.img', 'overlay-*']

        if self.clone:
            imgmnt = DiskMount(base_on, self._mkdtemp())
            ignore_list.remove('home.img')
        else:
            imgmnt = DiskMount(LoopbackDisk(base_on, 0), self._mkdtemp())
            print imgmnt.disk
            print imgmnt.mountdir
            print imgmnt.fstype
            print imgmnt.rmmountdir
            exit()

        self._LiveImageCreatorBase__isodir = self._ImageCreator__builddir + "/iso"

        try:
            imgmnt.mount()
            # legacy LiveOS filesystem layout support, remove for F9 or F10
            if os.path.exists(imgmnt.mountdir + "/squashfs.img"):
                squashimg = imgmnt.mountdir + "/squashfs.img"
            else:
                squashimg = imgmnt.mountdir + "/LiveOS/squashfs.img"

            # include specified files or directories
            if self._include:
                dst = self._LiveImageCreatorBase__isodir
                print self._include
                for fd in self._include:
                    if os.path.isfile(fd):
                        shutil.copy2(fd, os.path.join(dst, name))
                    elif os.path.isdir(fd):
                        shutil.copytree(fd,
                                        self._LiveImageCreatorBase__isodir)
            else:
                #copy over everything but squashimg
                shutil.copytree(imgmnt.mountdir,
                                self._LiveImageCreatorBase__isodir,
                                ignore=shutil.ignore_patterns(ignore_list)
        except MountError, e:
            raise CreatorError("Failed to mount '%s' : %s" %
                               (base_on, e))

        finally:
            imgmnt.cleanup()

    def _brand (self, _LiveImageCreatorBase__isodir, _instroot):

        # Get build name from boot configuration file.
        try:
            cfgf = open(_LiveImageCreatorBase__isodir
                        + "/syslinux/syslinux.cfg", "r")
                        
        except IOError:
            cfgf = open(_LiveImageCreatorBase__isodir
                + "/syslinux/extlinux.conf", "r")
        for line in cfgf:
            i = line.find("Welcome to ")
            if i > -1:
                self.name = line[i+11:-2]
                break
        cfgf.close()

        # Update fedora-release message with Remix details.
        try:
            text = open(_instroot + "/etc/fedora-release", "r").read()
        except IOError:
            raise CreatorError("Failed to open '%s' : %s" %
                               (cfgf, e))

        dt = datetime.strptime(datetime.now(), "%d-%b-%Y")
        text = text + "  Remix of %(name)s by %(_builder)s, packaged on %(dt)s"
        open(self._instroot + "/etc/fedora-release", "w").write(text)

    def __get_basic_syslinux_config(self, **args):
        dt = datetime.strptime(datetime.now(), "%d-%b-%Y")
        return """
default %(menu)s
timeout %(timeout)d

%(background)s
menu title Welcome to %(name)s %(dt)s Remix!
menu color border 0 #ffffffff #00000000
menu color sel 7 #ffffffff #ff000000
menu color title 0 #ffffffff #00000000
menu color tabmsg 0 #ffffffff #00000000
menu color unsel 0 #ffffffff #00000000
menu color hotsel 0 #ff000000 #ffffffff
menu color hotkey 7 #ffffffff #ff000000
menu color timeout_msg 0 #ffffffff #00000000
menu color timeout 0 #ffffffff #00000000
menu color cmdline 0 #ffffffff #00000000
menu hidden
menu hiddenrow 5
""" % args


def parse_options(args):
    parser = optparse.OptionParser(usage = "%prog [-n=<name>] \
                           \n\r                   [-o=<output>] \
                           \n\r                   [-s=<script.sh>] \
                           \n\r                   [-t=<tmpdir>] \
                           \n\r                   [-i=<includes>] \
                           \n\r                   [--builder] \
                           \n\r                   [--clone] \
                           \n\r                   [--skip-compression] \
                           \n\r                   [--skip-minimize] \
                           \n\r                   <LIVEIMG.src>")

    parser.add_option("-n", "--name", type="string", dest="name",
                      help="name of new LiveOS (don't include .iso, it will be added)")

    parser.add_option("-o", "--output", type="string", dest="output",
                      help="specify the output dir")

    parser.add_option("-s", "--script", type="string", dest="script",
                      help="specify script to run chrooted in the LiveOS fsimage")

    parser.add_option("", "--clone", action="store_true", dest="clone",
                      help="Specify that source image is LiveOS block device.")

    parser.add_option("-i", "--include", type="string", dest="include",
                      help="Specify directory or file patterns to be included in copy_img_root.")

    parser.add_option("", "--builder", type="string", dest="builder",
                      help="Specify the builder of a Remix.")

    parser.add_option("-t", "--tmpdir", type="string",
                      dest="tmpdir", default="/var/tmp",
                      help="Temporary directory to use (default: /var/tmp)")

    parser.add_option("", "--skip-compression", action="store_true", dest="skip_compression")

    parser.add_option("", "--skip-minimize", action="store_true", dest="skip_minimize")

    setup_logging(parser)

    (options, args) = parser.parse_args()

    if len(args) != 1:
        parser.print_usage()
        sys.exit(1)

    if stat.S_ISBLK(os.stat(args[0]).st_mode):
        options.clone = True

    return (args[0], options)

def rebuild_iso_symlinks(isodir):
    # remove duplicate files and rebuild symlinks to reduce iso size
    efi_vmlinuz = "%s/EFI/boot/vmlinuz0" % isodir
    isolinux_vmlinuz = "%s/isolinux/vmlinuz0" % isodir
    efi_initrd = "%s/EFI/boot/initrd0.img" % isodir
    isolinux_initrd = "%s/isolinux/initrd0.img" % isodir

    if os.path.exists(efi_vmlinuz):
        os.remove(efi_vmlinuz)
        os.remove(efi_initrd)
        os.symlink(isolinux_vmlinuz,efi_vmlinuz)
        os.symlink(isolinux_initrd,efi_initrd)


def main():
    (LiveOS, options) = parse_options(sys.argv[1:])

    if os.geteuid () != 0:
        print >> sys.stderr, "You must run edit-liveos as root"
        return 1

    if options.name and options.name != os.path.basename(LiveOS):
        name = options.name
    else:
        name = os.path.basename(LiveOS) + ".edited"

    if options.output:
        output = options.output
    else:
        output = os.path.dirname(LiveOS)

    editor = LiveImageEditor(name)
    editor._include = options.include
    editor.clone = options.clone
    editor.tmpdir = os.path.abspath(options.tmpdir)
    editor._builder = options.builder
    editor.skip_compression = options.skip_compression
    editor.skip_minimize = options.skip_minimize

    try:
        editor.mount(LiveOS, cachedir = None)
        if editor.clone:
            editor._brand(editor._LiveImageCreatorBase__isodir, self.__instloop)
        if options.script:
            print "Running edit script '%s'" % options.script
            editor._run_script(options.script)
        else:
            print "Launching shell. Exit to continue."
            print "----------------------------------"
            editor.launch_shell()
        rebuild_iso_symlinks(editor._LiveImageCreatorBase__isodir)
        editor.unmount()
        editor.package(output)
    except CreatorError, e:
        logging.error(u"Error editing LiveOS : %s" % e)
        return 1
    finally:
        editor.cleanup()

    return 0


if __name__ == "__main__":
    sys.exit(main())

arch = rpmUtils.arch.getBaseArch()
if arch in ("i386", "x86_64"):
    LiveImageCreator = x86LiveImageCreator
elif arch in ("ppc",):
    LiveImageCreator = ppcLiveImageCreator
elif arch in ("ppc64",):
    LiveImageCreator = ppc64LiveImageCreator
else:
    raise CreatorError("Architecture not supported!")
